#include <gtk/gtk.h>
#include <math.h>
#include <time.h>

#include "gwt_plot_canvas.h"
#include "gwt_plot_item.h"
#include "gwt_marshallers.h"

#define GWT_DRAG_MIN 50


#define GWT_PLOT_CANVAS_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GWT_TYPE_PLOT_CANVAS, GwtPlotCanvasPrivate))

G_DEFINE_TYPE (GwtPlotCanvas, gwt_plot_canvas, GTK_TYPE_DRAWING_AREA);


static gboolean gwt_plot_canvas_expose (GtkWidget *canvas, GdkEventExpose *event);
static gboolean gwt_plot_canvas_button_press (GtkWidget *canvas,
					     GdkEventButton *event);
static gboolean gwt_plot_canvas_button_release (GtkWidget *canvas,
					       GdkEventButton *event);
static gboolean gwt_plot_canvas_motion_notify (GtkWidget *canvas,
					      GdkEventMotion *event);
//static gboolean gwt_plot_canvas_update (gpointer data);
static gboolean gwt_plot_canvas_configure_event( GtkWidget *widget, GdkEventConfigure *event );
static void gwt_plot_canvas_dispose (GObject *obj);
static void gwt_plot_canvas_finalize (GObject *obj);
static void gwt_plot_canvas_size_request (GtkWidget        *canvas,
					 GtkRequisition   *requisition);
//static void gwt_plot_canvas_size_allocate (GtkWidget        *canvas,
//                                         GtkAllocation    *allocation);
static void gwt_plot_canvas_update_adjustments(GwtPlotCanvas *canvas);

static void gwt_plot_canvas_redraw_canvas (GtkWidget *canvas);

static gpointer    parent_class = NULL;

typedef struct _GwtPlotCanvasPrivate GwtPlotCanvasPrivate;

struct _GwtPlotCanvasPrivate
{
	GList *item_list;

	GdkPixmap *pixmap; /* backing store pixmap */

	gboolean dragging; /* true if the interface is being dragged */
	gint drag_start_x;
	gint drag_start_y;
	gint drag_end_x;
	gint drag_end_y;

	gint idle_function_id;
	
	gboolean need_redraw;
	gboolean dispose_has_run;
};

enum
{
	TIME_CHANGED,
	LAST_SIGNAL
};

//static guint gwt_plot_canvas_signals[LAST_SIGNAL] = { 0 };


static void
gwt_plot_canvas_unref_items_cb(gpointer data, gpointer user_data)
{
	g_debug("gwt_plot_canvas_unref_items_cb %p", data);
	g_object_unref (G_OBJECT(data));
}


static void
gwt_plot_canvas_dispose (GObject *obj)
{
	GwtPlotCanvasPrivate *priv;
	GwtPlotCanvas *canvas;

	canvas = GWT_PLOT_CANVAS(obj);
	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	if (priv->dispose_has_run) {
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	priv->dispose_has_run = TRUE;

	/* 
 	 * In dispose, you are supposed to free all types referenced from this
	 * object which might themselves hold a reference to self. Generally,
	 * the most simple solution is to unref all members on which you own a 
	 * reference.
	 */

	g_list_foreach (priv->item_list, gwt_plot_canvas_unref_items_cb, NULL);
	g_list_free(priv->item_list);
	priv->item_list = NULL;
	
	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static void
gwt_plot_canvas_finalize (GObject *obj)
{
	GwtPlotCanvasPrivate *priv;
	GwtPlotCanvas *canvas;

	canvas = GWT_PLOT_CANVAS(obj);

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->finalize (obj);
}



static void
gwt_plot_canvas_class_init (GwtPlotCanvasClass *class)
{
	GObjectClass *gobject_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS (class);
	widget_class = GTK_WIDGET_CLASS (class);

	parent_class = g_type_class_ref (G_TYPE_OBJECT);

	/* GObject signals */
	gobject_class->dispose = gwt_plot_canvas_dispose;
	gobject_class->finalize = gwt_plot_canvas_finalize;

	/* GtkWidget signals */
	widget_class->expose_event = gwt_plot_canvas_expose;
	widget_class->button_press_event = gwt_plot_canvas_button_press;
	widget_class->button_release_event = gwt_plot_canvas_button_release;
	widget_class->motion_notify_event = gwt_plot_canvas_motion_notify;
	widget_class->size_request = gwt_plot_canvas_size_request;
	//widget_class->size_allocate = gwt_plot_canvas_size_allocate;
	widget_class->configure_event = gwt_plot_canvas_configure_event;

#if 0
	/* GwtPlotCanvas signals */
	egg_clock_face_signals[TIME_CHANGED] = g_signal_new (
			"time-changed",
			G_OBJECT_CLASS_TYPE (obj_class),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET (EggClockFaceClass, time_changed),
			NULL, NULL,
			_clock_marshal_VOID__INT_INT,
			G_TYPE_NONE, 2,
			G_TYPE_INT,
			G_TYPE_INT);
#endif
	g_type_class_add_private (gobject_class, sizeof (GwtPlotCanvasPrivate));
}



static gboolean
gwt_plot_canvas_idle_cb (gpointer data)
{
	GwtPlotCanvas *canvas;
	GwtPlotCanvasPrivate *priv;

	canvas = GWT_PLOT_CANVAS (data);
	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);
	
	if (priv->need_redraw) {
		priv->need_redraw = FALSE;
		gwt_plot_canvas_redraw_canvas (GTK_WIDGET(canvas));
	}

	return FALSE; /* do not keep running this event */
}

static gboolean
gwt_plot_canvas_timer_cb (gpointer data)
{
	GwtPlotCanvas *canvas;
	GwtPlotCanvasPrivate *priv;

	canvas = GWT_PLOT_CANVAS (data);
	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);
	
	gtk_idle_add( gwt_plot_canvas_idle_cb, canvas);

	return FALSE; /* do not keep running this event */
}

static void 
gwt_plot_canvas_update(GwtPlotCanvas *canvas, guint delay_ms)
{
	GwtPlotCanvasPrivate *priv;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	if (delay_ms == 0) {
		priv->need_redraw = FALSE;
		gwt_plot_canvas_redraw_canvas (GTK_WIDGET(canvas));
	} else {
		if (priv->need_redraw == FALSE) {
			priv->need_redraw = TRUE;
			g_timeout_add(delay_ms, gwt_plot_canvas_idle_cb, canvas);

			//gtk_idle_add( gwt_plot_canvas_idle_cb, canvas);
		}
	}
}


static void
gwt_plot_canvas_init (GwtPlotCanvas *canvas)
{
	GwtPlotCanvasPrivate *priv;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	gtk_widget_add_events (GTK_WIDGET (canvas),
			GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK);


	priv->dispose_has_run = FALSE;
	//priv->need_redraw = TRUE;

	//priv->idle_function_id = gtk_idle_add( gwt_plot_canvas_idle_cb, canvas);


	gtk_widget_set_double_buffered(GTK_WIDGET(canvas), FALSE);

	gwt_plot_canvas_update (canvas, 0);

	/* update the clock once a second */
	// g_timeout_add (1000, gwt_plot_canvas_update, canvas);
}


/* Create a new backing pixmap of the appropriate size */
static gboolean
gwt_plot_canvas_configure_event( GtkWidget *widget, GdkEventConfigure *event )
{
	GwtPlotCanvasPrivate *priv;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (widget);

	if (priv->pixmap)
		g_object_unref(priv->pixmap);

	priv->pixmap = gdk_pixmap_new(widget->window,
				widget->allocation.width,
				widget->allocation.height,
				-1);

	 gdk_draw_rectangle (priv->pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);

	gwt_plot_canvas_redraw_canvas (widget);

	return TRUE;
}

static gboolean
gwt_plot_canvas_expose (GtkWidget *widget, GdkEventExpose *event)
{
	GwtPlotCanvasPrivate *priv;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (widget);

	//g_return_value_if_fail (priv != NULL, FALSE);

	gdk_draw_drawable(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			priv->pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

	return FALSE;
}



static void
gwt_plot_canvas_redraw_canvas (GtkWidget *widget)
{
	GwtPlotCanvasPrivate *priv;
	cairo_t *cr;
	GList* item;
	gdouble x1, x2, y1, y2, dx, dy;


	priv = GWT_PLOT_CANVAS_GET_PRIVATE (widget);

	if (priv->pixmap == NULL) return;

	/* get a cairo_t */
	cr = gdk_cairo_create (priv->pixmap);

	/* draw back ground */
	cairo_rectangle (cr, 0, 0, widget->allocation.width, widget->allocation.height);
	cairo_set_source_rgb (cr, 1, 1, 1);
	cairo_fill (cr);

	/* draw all the items */
	item = g_list_first(priv->item_list);
	while (item != NULL) {
		gwt_plot_item_draw( GWT_PLOT_ITEM(item->data), cr, widget->allocation.width, widget->allocation.height);
		item = g_list_next(item);
	}

	/* draw drag rectange */
	if (priv->dragging) {
		dx = fabs(priv->drag_end_x - priv->drag_start_x);
		dy = fabs(priv->drag_end_y - priv->drag_start_y);

		cairo_set_source_rgb (cr, 0, 0, 1);

		if (dx < GWT_DRAG_MIN && dy > GWT_DRAG_MIN) {
			cairo_move_to(cr, priv->drag_start_x, priv->drag_start_y);	
			cairo_line_to(cr, priv->drag_start_x, priv->drag_end_y);	
			cairo_stroke (cr);

			x1 = 0;
			x2 = widget->allocation.width;
			y1 = priv->drag_start_y;
			y2 = priv->drag_end_y;
		} else if (dx > GWT_DRAG_MIN && dy < GWT_DRAG_MIN) {
			cairo_move_to(cr, priv->drag_start_x, priv->drag_start_y);	
			cairo_line_to(cr, priv->drag_end_x, priv->drag_start_y);	
			cairo_stroke (cr);

			x1 = priv->drag_start_x;
			x2 = priv->drag_end_x;
			y1 = 0;
			y2 = widget->allocation.height;
		} else {
			x1 = priv->drag_start_x;
			x2 = priv->drag_end_x;
			y1 = priv->drag_start_y;
			y2 = priv->drag_end_y;
		}

	
		cairo_rectangle (cr, x1 + 0.5, y1 + 0.5, x2 - x1, y2 - y1);
		cairo_set_source_rgba (cr, 0, 0, 1, 0.2);
		cairo_fill_preserve (cr);
		cairo_set_source_rgb (cr, 0, 0, 1);
		cairo_stroke (cr);
	}

	/* draw border */
	cairo_set_line_width(cr, 1.0);
	cairo_rectangle (cr, 0.5, 0.5, widget->allocation.width - 1, widget->allocation.height - 1);
	cairo_set_source_rgb (cr, 0, 0, 0);
	cairo_stroke (cr);


	cairo_destroy (cr);

	gtk_widget_queue_draw_area( widget, 0, 0, widget->allocation.width, widget->allocation.height);
}

static gboolean
gwt_plot_canvas_button_press (GtkWidget *canvas, GdkEventButton *event)
{
	GwtPlotCanvasPrivate *priv;
	
	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	if (event->button == 1) {
		priv->drag_start_x = event->x;
		priv->drag_start_y = event->y;
		priv->drag_end_x = event->x;
		priv->drag_end_y = event->y;
		priv->dragging = TRUE;
	} else if (event->button == 3) {
		priv->drag_start_x = event->x - canvas->allocation.width;
		priv->drag_end_x = event->x + canvas->allocation.width;
		priv->drag_start_y = event->y - canvas->allocation.height;
		priv->drag_end_y = event->y + canvas->allocation.height;

		//priv->need_redraw = TRUE;

		gwt_plot_canvas_update(GWT_PLOT_CANVAS(canvas), 5);
		//gtk_widget_queue_draw_area( canvas, 0, 0, canvas->allocation.width, canvas->allocation.height);

		gwt_plot_canvas_update_adjustments(GWT_PLOT_CANVAS(canvas));
	}

	return FALSE;
}

static gboolean
gwt_plot_canvas_motion_notify (GtkWidget *canvas, GdkEventMotion *event)
{
	GwtPlotCanvasPrivate *priv;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	if (priv->dragging)
	{
		priv->drag_end_x = event->x;
		priv->drag_end_y = event->y;
	
		//priv->need_redraw = TRUE;			
		//gtk_widget_queue_draw_area( canvas, 0, 0, canvas->allocation.width, canvas->allocation.height);
		gwt_plot_canvas_update(GWT_PLOT_CANVAS(canvas), 5);

		//emit_time_changed_signal (EGG_CLOCK_FACE (clock),
		//		event->x, event->y);
	}

	return FALSE;
}

static gboolean
gwt_plot_canvas_button_release (GtkWidget *canvas, GdkEventButton *event)
{
	GwtPlotCanvasPrivate *priv;
	gdouble dx, dy , tmp;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);
	

	if (priv->dragging)
	{
		priv->dragging = FALSE;

		priv->drag_end_x = event->x;
		priv->drag_end_y = event->y;

		if ((priv->drag_end_x == priv->drag_start_x) || (priv->drag_end_y == priv->drag_start_y))
			return FALSE;

		//priv->need_redraw = TRUE;
		gwt_plot_canvas_update(GWT_PLOT_CANVAS(canvas), 5);


		dx = fabs(priv->drag_end_x - priv->drag_start_x);
		dy = fabs(priv->drag_end_y - priv->drag_start_y);

		if (dx < GWT_DRAG_MIN && dy > GWT_DRAG_MIN) {
			priv->drag_start_x = 0;
			priv->drag_end_x = GTK_WIDGET(canvas)->allocation.width;
		} else if (dx > GWT_DRAG_MIN && dy < GWT_DRAG_MIN) {
			priv->drag_start_y = 0;
			priv->drag_end_y = GTK_WIDGET(canvas)->allocation.height;
		}

		if (priv->drag_start_x > priv->drag_end_x) {
			tmp = priv->drag_start_x;
			priv->drag_start_x = priv->drag_end_x;
			priv->drag_end_x = tmp;
		}

		if (priv->drag_start_y > priv->drag_end_y) {
			tmp = priv->drag_start_y;
			priv->drag_start_y = priv->drag_end_y;
			priv->drag_end_y = tmp;
		}

		if (dx > GWT_DRAG_MIN || dy > GWT_DRAG_MIN)
			gwt_plot_canvas_update_adjustments(GWT_PLOT_CANVAS(canvas));
	}

	return FALSE;
}

static void
gwt_plot_canvas_size_request (GtkWidget      *widget,
			GtkRequisition *requisition)
{
	GwtPlotCanvas *canvas;
	GwtPlotCanvasPrivate *priv;
  
	g_return_if_fail (GWT_IS_PLOT_CANVAS (widget));
	g_return_if_fail (requisition != NULL);
  
	canvas = GWT_PLOT_CANVAS (widget);
	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	requisition->width = 200;
	requisition->height = 200;
}

static void
gwt_plot_canvas_size_allocate (GtkWidget     *widget,
                         GtkAllocation *allocation)
{
	GwtPlotCanvas *canvas;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GWT_IS_PLOT_CANVAS (widget));
	g_return_if_fail (allocation != NULL);

	canvas = GWT_PLOT_CANVAS (widget);

	//(* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);


	widget->allocation = *allocation;
	if (GTK_WIDGET_REALIZED (widget)) {
		gdk_window_move_resize (widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);

		//dial->radius = MAX(allocation->width,allocation->height) * 0.45;
		//dial->pointer_width = dial->radius / 5;
	}
}


struct adjustment_info {
	GTree *x_adjustments;
	GTree *y_adjustments;
};

static void
gwt_plot_canvas_select_adjustments_cb(gpointer data, gpointer user_data)
{
	struct adjustment_info *ai = (struct adjustment_info*)user_data;
	GwtPlotItem *item = GWT_PLOT_ITEM(data);
	GwtPlotAdjustment* adj;

	adj = gwt_plot_item_get_x_adjustment(item);
	g_tree_insert (ai->x_adjustments, adj, NULL);

	adj = gwt_plot_item_get_y_adjustment(item);
	g_tree_insert (ai->y_adjustments, adj, NULL);
}

static gboolean
gwt_plot_canvas_update_x_adjustments_cb(gpointer key, gpointer value, gpointer data)
{
	gdouble scale, start_value, end_value, tmp;

	GwtPlotCanvas* canvas = GWT_PLOT_CANVAS(data);
	GwtPlotAdjustment* adj = GWT_PLOT_ADJUSTMENT(key);
	GwtPlotCanvasPrivate *priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);
	

	start_value = gwt_plot_adjustment_get_start_value( adj );
	end_value = gwt_plot_adjustment_get_end_value( adj );

	scale = (end_value - start_value) / GTK_WIDGET(canvas)->allocation.width;

	end_value = (((gdouble)priv->drag_end_x) * scale) + start_value;
	start_value = (((gdouble)priv->drag_start_x) * scale) + start_value;

	if (end_value < start_value) {
		tmp = end_value;
		end_value = start_value;
		start_value = tmp;
	}

	gwt_plot_adjustment_zoom( adj, start_value, end_value);

	return FALSE;
}

static gboolean
gwt_plot_canvas_update_y_adjustments_cb(gpointer key, gpointer value, gpointer data)
{
	gdouble scale, start_value, end_value, tmp;

	GwtPlotCanvas* canvas = GWT_PLOT_CANVAS(data);
	GwtPlotAdjustment* adj = GWT_PLOT_ADJUSTMENT(key);
	GwtPlotCanvasPrivate *priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	start_value = gwt_plot_adjustment_get_start_value( adj );
	end_value = gwt_plot_adjustment_get_end_value( adj );

	scale = (end_value - start_value) / GTK_WIDGET(canvas)->allocation.height;

	end_value = (((gdouble)GTK_WIDGET(canvas)->allocation.height - priv->drag_end_y) * scale) + start_value;
	start_value = (((gdouble)GTK_WIDGET(canvas)->allocation.height - priv->drag_start_y) * scale) + start_value;

	if (end_value < start_value) {
		tmp = end_value;
		end_value = start_value;
		start_value = tmp;
	}

	gwt_plot_adjustment_zoom( adj, start_value, end_value);

	return FALSE;
}

static gint
gwt_plot_canvas_pointer_cmp (gconstpointer a, gconstpointer b)
{
	if (a < b)
		return -1;
	else if (a > b)
		return 1;
	else
		return 0;
}

static void
gwt_plot_canvas_update_adjustments(GwtPlotCanvas *canvas) 
{
	GwtPlotCanvasPrivate *priv;
	struct adjustment_info ai;

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	ai.x_adjustments = g_tree_new(gwt_plot_canvas_pointer_cmp);
	ai.y_adjustments = g_tree_new(gwt_plot_canvas_pointer_cmp);

	g_list_foreach (priv->item_list, gwt_plot_canvas_select_adjustments_cb, &ai);

	g_tree_foreach (ai.x_adjustments, gwt_plot_canvas_update_x_adjustments_cb, canvas);
	g_tree_foreach (ai.y_adjustments, gwt_plot_canvas_update_y_adjustments_cb, canvas);

	g_tree_destroy (ai.x_adjustments);
	g_tree_destroy (ai.y_adjustments);
}

void
gwt_plot_canvas_item_changed_cb(GwtPlotItem* item, GwtPlotCanvas* canvas)
{
	GwtPlotCanvasPrivate *priv;

	g_return_if_fail (canvas != NULL);
	g_return_if_fail (GWT_IS_PLOT_CANVAS (canvas));
	g_return_if_fail (item != NULL);
	g_return_if_fail (GWT_IS_PLOT_ITEM (item));

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (canvas);

	//priv->need_redraw = TRUE;
	
	gwt_plot_canvas_update (canvas, 50);
}

GtkWidget *
gwt_plot_canvas_new (void)
{
	return g_object_new (GWT_TYPE_PLOT_CANVAS, NULL);
}

void 
gwt_plot_canvas_add_item(GwtPlotCanvas* widget, GwtPlotItem* item)
{
	GwtPlotCanvasPrivate *priv;
	GwtPlotAdjustment* adj;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GWT_IS_PLOT_CANVAS (widget));
	g_return_if_fail (item != NULL);
	g_return_if_fail (GWT_IS_PLOT_ITEM (item));

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (widget);

	g_object_ref_sink (item);
	priv->item_list = g_list_append( priv->item_list, item);

	g_signal_connect (item, "changed", G_CALLBACK (gwt_plot_canvas_item_changed_cb), widget);

	gwt_plot_canvas_update (widget, 0);	
}

void 
gwt_plot_canvas_remove_item(GwtPlotCanvas* widget, GwtPlotItem* item)
{
	GwtPlotCanvasPrivate *priv;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GWT_IS_PLOT_CANVAS (widget));
	g_return_if_fail (item != NULL);
	g_return_if_fail (GWT_IS_PLOT_ITEM (item));

	priv = GWT_PLOT_CANVAS_GET_PRIVATE (widget);

	g_signal_handlers_disconnect_by_func(item, G_CALLBACK (gwt_plot_canvas_item_changed_cb), widget);

	priv->item_list = g_list_remove(priv->item_list, item);
	g_object_unref (item);

	gwt_plot_canvas_update (widget, 0);	
}

