#include <gtk/gtk.h>
#include <math.h>
#include <time.h>

#include "gwt_marshallers.h"
#include "gwt_plot_curve.h"
#include "gwt_plot_item.h"
#include "gwt_plot_data.h"

#define GWT_PLOT_CURVE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GWT_TYPE_PLOT_CURVE, GwtPlotCurvePrivate))


G_DEFINE_TYPE (GwtPlotCurve, gwt_plot_curve, GWT_TYPE_PLOT_ITEM)


typedef struct _GwtPlotCurvePrivate GwtPlotCurvePrivate;

struct _GwtPlotCurvePrivate
{	
	GwtPlotData* data;

	GwtPlotCurveStyle style;
	GwtPlotCurveSymbol symbol;
	GwtPlotCurveLineStyle line_style;
};

enum
{
	CHANGED,
	LAST_SIGNAL
};

struct sample {
	gdouble x;
	gdouble y_min;
	gdouble y_avg;
	gdouble y_max;
	gdouble y2; // for spline
};

static guint gwt_plot_curve_signals[LAST_SIGNAL] = { 0 };



static int
project (gfloat value, gfloat min, gfloat max, int norm)
{
	return (norm - 1) * ((value - min) / (max - min)) + 0.5;
}

static gfloat
unproject (gint value, gfloat min, gfloat max, int norm)
{
	return value / (gfloat) (norm - 1) * (max - min) + min;
}

/* Solve the tridiagonal equation system that determines the second
   derivatives for the interpolation points.  (Based on Numerical
   Recipies 2nd Edition.) */
static void
spline_solve (int n, struct sample* sample)
{
	gdouble p, sig, *u;
	gint i, k;

	u = g_malloc ((n - 1) * sizeof (u[0]));

	sample[0].y2 = u[0] = 0.0;	/* set lower boundary condition to "natural" */

	for (i = 1; i < n - 1; ++i)
	{
		sig = (sample[i].x - sample[i - 1].x) / (sample[i + 1].x - sample[i - 1].x);
		p = sig * sample[i - 1].y2 + 2.0;
		sample[i].y2 = (sig - 1.0) / p;
		u[i] = ((sample[i + 1].y_avg - sample[i].y_avg)
			/ (sample[i + 1].x - sample[i].x) - (sample[i].y_avg - sample[i - 1].y_avg) / (sample[i].x - sample[i - 1].x));
		u[i] = (6.0 * u[i] / (sample[i + 1].x - sample[i - 1].x) - sig * u[i - 1]) / p;
	}

	sample[n - 1].y2 = 0.0;
	for (k = n - 2; k >= 0; --k)
		sample[k].y2 = sample[k].y2 * sample[k + 1].y2 + u[k];

	g_free (u);
}

static gdouble
spline_eval (int n, struct sample* sample, gdouble val)
{
	gint k_lo, k_hi, k;
	gdouble h, b, a;

	if ((n < 3) || (val < sample[0].x) || (val >= sample[n-1].x))
		return G_MAXDOUBLE;

	/* do a binary search for the right interval: */
	k_lo = 0; k_hi = n - 1;
	while (k_hi - k_lo > 1)
	{
		k = (k_hi + k_lo) / 2;
		if (sample[k].x > val)
			k_hi = k;
		else
			k_lo = k;
	}

	h = sample[k_hi].x - sample[k_lo].x;
	g_assert (h > 0.0);

	a = (sample[k_hi].x - val) / h;
	b = (val - sample[k_lo].x) / h;

	return a*sample[k_lo].y_avg + b*sample[k_hi].y_avg +
		((a*a*a - a)*sample[k_lo].y2 + (b*b*b - b)*sample[k_hi].y2) * (h*h)/6.0;
}

static void
gwt_plot_curve_draw_symbol(GwtPlotCurve* curve, cairo_t *cr, gdouble x, gdouble y)
{
	GwtPlotCurvePrivate *priv;
	gdouble dx, dy;

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	dx = 4.0;
	dy = 4.0;

	cairo_set_source_rgb (cr, GWT_PLOT_ITEM(curve)->color[0], GWT_PLOT_ITEM(curve)->color[1], GWT_PLOT_ITEM(curve)->color[2]);

	switch (priv->symbol) {
	case GWT_PLOT_CURVE_SYMBOL_CROSS:
		cairo_move_to(cr, x - dx, y);
		cairo_line_to(cr, x + dx, y);
		cairo_move_to(cr, x, y - dy);
		cairo_line_to(cr, x, y + dy);
		cairo_stroke(cr);
		break;

	case GWT_PLOT_CURVE_SYMBOL_XCROSS:
		dx -= 1.0;
		dy -= 1.0;
		cairo_move_to(cr, x - dx, y + dy);
		cairo_line_to(cr, x + dx, y - dy);
		cairo_move_to(cr, x - dx, y - dy);
		cairo_line_to(cr, x + dx, y + dy);
		cairo_stroke(cr);
		break;

	case GWT_PLOT_CURVE_SYMBOL_DOT:
		cairo_arc (cr, x, y, dx, 0, 2 * M_PI);
		cairo_fill(cr);
		break;

	case GWT_PLOT_CURVE_SYMBOL_SQUARE:
		cairo_move_to(cr, x - dx, y + dy);
		cairo_line_to(cr, x + dx, y + dy);
		cairo_line_to(cr, x + dx, y - dy);
		cairo_line_to(cr, x - dx, y - dy);
		cairo_fill(cr);
		break;	

	case GWT_PLOT_CURVE_SYMBOL_DIAMOND:
		dx += 1;
		dy += 1;
		cairo_move_to(cr, x , y + dy);
		cairo_line_to(cr, x + dx, y);
		cairo_line_to(cr, x, y - dy);
		cairo_line_to(cr, x - dx, y);
		cairo_fill(cr);
		break;
	
	case GWT_PLOT_CURVE_SYMBOL_NONE:
	default:
		break;
	};
}

static void
gwt_plot_curve_draw_line(GwtPlotCurve* curve, cairo_t* cr, struct sample* sample, guint sample_count, gdouble dev_x_0, gdouble dev_y_0, int w, int h) {
	GwtPlotCurvePrivate *priv;
	int i;
	gdouble y;
	static gdouble dot_patern[2] = { 3.0, 3.0 };
	static gdouble dash_patern[2] = { 10.0, 3.0 };
	static gdouble dot_dash_patern[4] = { 3.0, 3.0, 10.0, 3.0};


  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	/* draw line's */
	cairo_save(cr);

	cairo_set_source_rgb (cr,  GWT_PLOT_ITEM(curve)->color[0], GWT_PLOT_ITEM(curve)->color[1], GWT_PLOT_ITEM(curve)->color[2]);
	cairo_set_line_width (cr, 1.0);
	//cairo_set_antialias (cr , CAIRO_ANTIALIAS_NONE);

	switch (priv->line_style) {
	case GWT_PLOT_CURVE_LINE_STYLE_DOT:
		cairo_set_dash(cr, dot_patern, 2, 0);
		break;

	case GWT_PLOT_CURVE_LINE_STYLE_DOT_DASH:
		cairo_set_dash(cr, dot_dash_patern, 4, 0);
		break;

	case GWT_PLOT_CURVE_LINE_STYLE_DASH:
		cairo_set_dash(cr, dash_patern, 2, 0);
		break;
	
	case GWT_PLOT_CURVE_LINE_STYLE_NORMAL:
	default:
		break;
	};

	switch (priv->style) {
	case GWT_PLOT_CURVE_STYLE_NONE:
		break;
	case GWT_PLOT_CURVE_STYLE_FITTED:
		spline_solve(sample_count, sample);

		i = 0;
		while ((i <= w) && (spline_eval(sample_count, sample, i++) == G_MAXDOUBLE)) ;

		cairo_move_to(cr, i, spline_eval(sample_count, sample, i));
		while (i++ <= w) {
			y = spline_eval(sample_count, sample, i);
			if (y == G_MAXDOUBLE)
				break;
			cairo_line_to (cr, i, y);
		}
		cairo_stroke (cr);

		break;
	case GWT_PLOT_CURVE_STYLE_LINES:
		for (i = 0; i < sample_count; i++) {
			if (sample[i].x < 0)
				cairo_move_to(cr , sample[i].x, sample[i].y_avg);
			else 
				cairo_line_to (cr, sample[i].x, sample[i].y_avg);
		}
		cairo_stroke (cr);
		break;
		
	case GWT_PLOT_CURVE_STYLE_STEPS:
		cairo_move_to(cr , sample[0].x, sample[0].y_avg);
		for (i = 1; i < sample_count; i++) {
			if (sample[i].x < 0) {
				cairo_move_to(cr , sample[i].x, sample[i].y_avg);
			} else {
				cairo_line_to (cr, sample[i].x, sample[i-1].y_avg);
				cairo_line_to (cr, sample[i].x, sample[i].y_avg);
			}
		}
		cairo_stroke (cr);
		break;

	case GWT_PLOT_CURVE_STYLE_STICKS:
		for (i = 0; i < sample_count; i++) {
			if (sample[i].x >= 0) {
				cairo_move_to (cr, sample[i].x, dev_y_0);
				cairo_line_to (cr, sample[i].x, sample[i].y_avg);
			}
		}
		cairo_stroke (cr);
		break;
	};

	cairo_restore(cr);
}


static void
gwt_plot_curve_draw_impl (GwtPlotItem* item, cairo_t *cr, int w, int h)
{	
	GwtPlotCurvePrivate *priv;
	GwtPlotAdjustment *x_adj, *y_adj;
	gdouble x_min, x_max, y_min, y_max;
	gdouble dx, dy, prev_x;
	GwtPlotDataPoint data_point;
	guint index, start_index;
	struct sample* sample;
	guint avg_count;
	gdouble dev_x_0, dev_y_0;
	
	gdouble x;
	guint i;
	guint point_count;

	g_return_if_fail (item != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (item));
	g_return_if_fail (cr != NULL);

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (item);

	if (!GWT_PLOT_ITEM(item)->is_visible)
		return;

	if (gwt_plot_data_get(priv->data, 5, &data_point) == -1)
		return;
		
	x_adj = gwt_plot_item_get_x_adjustment (item);

	g_return_if_fail (x_adj != NULL);
	gwt_plot_adjustment_set_device_interval(x_adj, 0, w);
	x_min = gwt_plot_adjustment_get_start_value(x_adj);
	x_max = gwt_plot_adjustment_get_end_value(x_adj);


	y_adj = gwt_plot_item_get_y_adjustment (item);

	g_return_if_fail (y_adj != NULL);
	gwt_plot_adjustment_set_device_interval(y_adj, 0, h);
	y_min = gwt_plot_adjustment_get_start_value(y_adj);
	y_max = gwt_plot_adjustment_get_end_value(y_adj);

	/* room for the whole screen plus half the screen on both sides */
	sample = g_new(struct sample, w * 2);	

	/* get sample points and scale them to device units */

	dev_y_0 = (gdouble)h - gwt_plot_adjustment_user_to_device(y_adj, 0.0);
	dev_x_0 = gwt_plot_adjustment_user_to_device(x_adj, 0.0);

	if ((index = gwt_plot_data_find_smaller ( priv->data, x_min, &data_point)) == -1)
		index = 0;


	data_point.y = (gdouble)h - gwt_plot_adjustment_user_to_device(y_adj, data_point.y);	
	data_point.x = gwt_plot_adjustment_user_to_device(x_adj, data_point.x);

	/* we found the first index, now go back atleast 5 samples */
	avg_count = 1;
	i = 0;	
	x = round(data_point.x);
	while (gwt_plot_data_get(priv->data, index--, &data_point) != -1) {
		data_point.y = (gdouble)h - gwt_plot_adjustment_user_to_device(y_adj, data_point.y);	
		data_point.x = gwt_plot_adjustment_user_to_device(x_adj, data_point.x);

		if (round(data_point.x) != x) {
			i++;
			if (i == 5)
				break;
			x = round(data_point.x);
		} 
	}

	if ((index = gwt_plot_data_get(priv->data, index, &data_point)) == -1)
		index = 0;

	gwt_plot_data_get(priv->data, index, &data_point);

	data_point.y = (gdouble)h - gwt_plot_adjustment_user_to_device(y_adj, data_point.y);	
	data_point.x = gwt_plot_adjustment_user_to_device(x_adj, data_point.x);

	avg_count = 1;
	i = 0;	
	point_count = 0;
	sample[i].x = round(data_point.x);
	sample[i].y_min = sample[i].y_max = sample[i].y_avg = data_point.y;
	while (gwt_plot_data_get(priv->data, index++, &data_point) != -1) {
		data_point.y = (gdouble)h - gwt_plot_adjustment_user_to_device(y_adj, data_point.y);	
		data_point.x = gwt_plot_adjustment_user_to_device(x_adj, data_point.x);

		x = round(data_point.x);

		if (x == sample[i].x) {
			if (sample[i].y_min > data_point.y)
				sample[i].y_min = data_point.y;
			if (sample[i].y_max < data_point.y)
				sample[i].y_max = data_point.y;

			sample[i].y_avg += data_point.y;
			avg_count++;
		} else {
			sample[i].y_avg /= (gdouble)avg_count;
			avg_count = 1;

			/* make sure we don't go out of bounds on our array */
			if (i >= ((w * 2) - 1))
				break;

			if (sample[i].x < w) {
				point_count = i;
			}
		
			/* check if we are 5 samples past the end of the screen */
			if ((i - point_count) > 5)
				break;

			i++;
			sample[i].x = x;
			sample[i].y_min = data_point.y;
			sample[i].y_max = data_point.y;
			sample[i].y_avg = data_point.y;

		}
	}

	if (index == -1) {
		sample[i].y_avg /= (gdouble)avg_count;
	}

	point_count = i + 1;
	
	/* draw lines */
	if (priv->style != GWT_PLOT_CURVE_STYLE_NONE) {
		cairo_save(cr);
		gwt_plot_curve_draw_line(item, cr, sample, point_count, dev_x_0, dev_y_0, w, h);
		cairo_restore(cr);
	}

	/* only draw symbols if needed */
	if (priv->symbol != GWT_PLOT_CURVE_SYMBOL_NONE) {
		cairo_save(cr);

		cairo_set_antialias (cr , CAIRO_ANTIALIAS_NONE);

		for (i = 0; i < point_count; i++) {
			if (sample[i].x > 0 && sample[i].x < w) {
				if (sample[i].y_min != sample[i].y_max) {
					gwt_plot_curve_draw_symbol(item, cr, sample[i].x, sample[i].y_min);
					gwt_plot_curve_draw_symbol(item, cr, sample[i].x, sample[i].y_max);
				}
			
				gwt_plot_curve_draw_symbol(item, cr, sample[i].x, sample[i].y_avg);
			}
		}
	
		cairo_restore(cr);
	}


	g_free(sample);
}

static void
gwt_plot_curve_class_init (GwtPlotCurveClass *class)
{
	GObjectClass *obj_class;
	GwtPlotItemClass *item_class;
	//GtkWidgetClass *widget_class;

	obj_class = G_OBJECT_CLASS (class);
	item_class = GWT_PLOT_ITEM_CLASS(class);
	//widget_class = GTK_WIDGET_CLASS (class);

	/* GtkWidget signals */
	//widget_class->expose_event = gwt_plot_canvas_expose;
	//widget_class->button_press_event = gwt_plot_canvas_button_press;
	//widget_class->button_release_event = gwt_plot_canvas_button_release;
	//widget_class->motion_notify_event = gwt_plot_canvas_motion_notify;

	/* virtual functions */
	item_class->draw = gwt_plot_curve_draw_impl;


#if 0
	/* GwtPlotItem signals */
	gwt_plot_item_signals[CHANGED] =
		g_signal_new (
			"changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
			G_STRUCT_OFFSET (GwtPlotItemClass, changed),
			NULL, NULL,
			_gwt_marshal_VOID__VOID,
			G_TYPE_NONE, 0);

	gwt_plot_item_signals[VALUE_CHANGED] =
		g_signal_new (
			"value_changed",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
			G_STRUCT_OFFSET (GwtPlotItemClass, value_changed),
			NULL, NULL,
			_gwt_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
#endif

	g_type_class_add_private (obj_class, sizeof (GwtPlotCurvePrivate));
}

static void
gwt_plot_curve_init (GwtPlotCurve *curve)
{
}

static void
data_changed_cb(GwtPlotData* data, GwtPlotCurve* curve)
{
	//g_debug("data_changed_cb %p %p", data, curve);

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));
	g_return_if_fail (data != NULL);
	g_return_if_fail (GWT_IS_PLOT_DATA (data));

	gwt_plot_item_emit_changed(GWT_PLOT_ITEM(curve));
}

static void
data_value_changed_cb(GwtPlotData* data, double x, double y, GwtPlotCurve* curve)
{
	//g_debug("data_value_changed_cb %p %p %f %f",data, curve, x, y);

	GwtPlotCurvePrivate *priv;
	GwtPlotAdjustment* adj;
	gdouble x_min, x_max;

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));
	g_return_if_fail (data != NULL);
	g_return_if_fail (GWT_IS_PLOT_DATA (data));

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	adj = gwt_plot_item_get_x_adjustment (GWT_PLOT_ITEM(curve));

	g_return_if_fail (adj != NULL);

	x_min = gwt_plot_adjustment_get_start_value(adj);
	x_max = gwt_plot_adjustment_get_end_value(adj);

	if ((x >= x_min) && (x <= x_max))
		gwt_plot_item_emit_changed(GWT_PLOT_ITEM(curve));
}

GtkObject *
gwt_plot_curve_new (void)
{
	return g_object_new (GWT_TYPE_PLOT_CURVE, NULL);
}

void 
gwt_plot_curve_set_data (GwtPlotCurve *curve, GwtPlotData* data)
{
	GwtPlotCurvePrivate *priv;

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));
	g_return_if_fail (data != NULL);
	g_return_if_fail (GWT_IS_PLOT_DATA (data));

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	if (priv->data != NULL) {
		/* remove signal handlers */
		g_signal_handlers_disconnect_by_func(priv->data, G_CALLBACK (data_changed_cb), curve);
		g_signal_handlers_disconnect_by_func(priv->data, G_CALLBACK (data_value_changed_cb), curve);
	
		/* remove the old data */
		g_object_unref(priv->data);
	}

	priv->data = data;
	g_object_ref_sink (priv->data);

	/* set the signal handler */
	g_signal_connect (data, "changed", G_CALLBACK (data_changed_cb), curve);
	g_signal_connect (data, "value_changed", G_CALLBACK (data_value_changed_cb), curve);
}

void gwt_plot_curve_set_style(GwtPlotCurve *curve, GwtPlotCurveStyle style)
{
	GwtPlotCurvePrivate *priv;

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	if (priv->style != style) {
		priv->style = style;
		gwt_plot_item_emit_changed(GWT_PLOT_ITEM(curve));
	}
}

void gwt_plot_curve_set_symbol(GwtPlotCurve *curve, GwtPlotCurveSymbol symbol)
{
	GwtPlotCurvePrivate *priv;

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	if (priv->symbol != symbol ) {
		priv->symbol = symbol;
		gwt_plot_item_emit_changed(GWT_PLOT_ITEM(curve));
	}
}

void gwt_plot_curve_set_line_style(GwtPlotCurve *curve, GwtPlotCurveLineStyle line_style)
{
	GwtPlotCurvePrivate *priv;

	g_return_if_fail (curve != NULL);
	g_return_if_fail (GWT_IS_PLOT_CURVE (curve));

  	priv = GWT_PLOT_CURVE_GET_PRIVATE (curve);

	if (priv->line_style != line_style) {
		priv->line_style = line_style;
		gwt_plot_item_emit_changed(GWT_PLOT_ITEM(curve));
	}
}

