#include <gtk/gtk.h>
#include <math.h>
#include <time.h>
#include <string.h>

#include "gwt_plot_scale.h"
#include "gwt_plot_item.h"
#include "gwt_marshallers.h"
#include "gwt_typebuiltins.h"

#define GWT_PLOT_SCALE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GWT_TYPE_PLOT_SCALE, GwtPlotScalePrivate))

G_DEFINE_TYPE (GwtPlotScale, gwt_plot_scale, GTK_TYPE_DRAWING_AREA);


static gboolean gwt_plot_scale_expose (GtkWidget *scale, GdkEventExpose *event);
static gboolean gwt_plot_scale_button_press (GtkWidget *scale,
					     GdkEventButton *event);
static gboolean gwt_plot_scale_button_release (GtkWidget *scale,
					       GdkEventButton *event);
static gboolean gwt_plot_scale_motion_notify (GtkWidget *scale,
					      GdkEventMotion *event);
static gboolean gwt_plot_scale_scroll_event (GtkWidget *scale,
					     GdkEventScroll *event);
static gboolean gwt_plot_scale_update (gpointer data);

static void gwt_plot_scale_dispose (GObject *obj);
static void gwt_plot_scale_finalize (GObject *obj);
static void gwt_plot_scale_size_request (GtkWidget        *canvas,
					 GtkRequisition   *requisition);
static void gwt_plot_scale_size_allocate (GtkWidget        *canvas,
                                         GtkAllocation    *allocation);

static gpointer    parent_class = NULL;

typedef struct _GwtPlotScalePrivate GwtPlotScalePrivate;

struct _GwtPlotScalePrivate
{
	GwtPlotScaleType scale_type;
	gdouble tick_base;
	guint minor_ticks;

	gboolean dragging; /* true if the interface is being dragged */
	gdouble drag_start_x;
	gdouble drag_start_y;
	gboolean dispose_has_run;


	guint gap_size;		/* size of the gab between the side and the ticks */
	guint tick_size;	/* size of the major ticks */
	guint text_size;	/* size of the text */
	guint label_size;	/* size of the label */

	char* label;

	gdouble label_color[3];	/* the label RGB color */
	gdouble text_color[3];	/* the text RGB color */
	gdouble tick_color[3];	/* the tick RGB Color */
	gdouble bg_color[3];	/* the background RGB color */

	/* the adjustment that controles this scale */
	GwtPlotAdjustment* adjustment;
	GwtFormatCallback format_callback;
};

enum
{
	PROP_0,
	PROP_SCALE_TYPE,
};


enum
{
	TIME_CHANGED,
	LAST_SIGNAL
};

//static guint gwt_plot_canvas_signals[LAST_SIGNAL] = { 0 };

static void gwt_plot_scale_get_property (GObject      *object,
                                         guint         prop_id,
                                         GValue       *value,
                                         GParamSpec   *pspec);
static void gwt_plot_scale_set_property (GObject      *object,
                                         guint         prop_id,
                                         const GValue *value,
                                         GParamSpec   *pspec);

static gint default_format_callback(GwtPlotScale* scale, gdouble value, char* buffer, guint buffer_size)
{
	return snprintf(buffer, buffer_size, "%g", value);
}

static const double _eps = 1.0e-6;

static int compare_eps(double value1, double value2, double intervalSize) 
{
	double eps = fabs(_eps * intervalSize);

	if ( value2 - value1 > eps )
		return -1;

	if ( value1 - value2 > eps )
		return 1;

	return 0;
}

static double ceil_eps(double value,  double intervalSize) 
{
	double eps = _eps * intervalSize;

	value = (value - eps) / intervalSize;
	return ceil(value) * intervalSize;
}

static double floor_eps(double value, double intervalSize) 
{
	double eps = _eps * intervalSize;

	value = (value + eps) / intervalSize;
	return floor(value) * intervalSize;
}

/*
  \brief Divide an interval into steps
 
  \f$stepSize = (intervalSize - intervalSize * 10e^{-6}) / numSteps\f$

  \param intervalSize Interval size
  \param numSteps Number of steps
  \return Step size
*/
static double divide_eps(double intervalSize, double numSteps) 
{
	if ( numSteps == 0.0 || intervalSize == 0.0 )
		return 0.0;
 
	return (intervalSize - (_eps * intervalSize)) / numSteps;
} 

static double ceil125(double x) 
{
	if (x == 0.0) 
		return 0.0;
 
	double sign = (x > 0) ? 1.0 : -1.0;
	double lx = log10(fabs(x));
	double p10 = floor(lx);
    
	double fr = pow(10.0, lx - p10);
	if (fr <=1.0)
		fr = 1.0; 
	else if (fr <= 2.0)
		fr = 2.0;
	else if (fr <= 5.0) 
		fr = 5.0;
	else
		fr = 10.0;
 
	return sign * fr * pow(10.0, p10);
}

static double floor125(double x) 
{
	if (x == 0.0)
		return 0.0;
 
	double sign = (x > 0) ? 1.0 : -1.0;
	double lx = log10(fabs(x));
	double p10 = floor(lx);
 	double fr = pow(10.0, lx - p10);

	if (fr >= 10.0)
		fr = 10.0;
	else if (fr >= 5.0)
		fr = 5.0;
	else if (fr >= 2.0)
		fr = 2.0;
	else
		fr = 1.0;
 
	return sign * fr * pow(10.0, p10);
}

static void
gwt_plot_scale_dispose (GObject *obj)
{
	GwtPlotScalePrivate *priv;
	GwtPlotScale *scale;

	scale = GWT_PLOT_SCALE(obj);
	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	if (priv->dispose_has_run) {
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	priv->dispose_has_run = TRUE;

	/* 
 	 * In dispose, you are supposed to free all types referenced from this
	 * object which might themselves hold a reference to self. Generally,
	 * the most simple solution is to unref all members on which you own a 
	 * reference.
	 */


	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static void
gwt_plot_scale_finalize (GObject *obj)
{
	GwtPlotScalePrivate *priv;
	GwtPlotScale *scale;

	scale = GWT_PLOT_SCALE(obj);

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);


	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->finalize (obj);
}



static void
gwt_plot_scale_class_init (GwtPlotScaleClass *class)
{
	GObjectClass *gobject_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS (class);
	widget_class = GTK_WIDGET_CLASS (class);

	parent_class = g_type_class_ref (G_TYPE_OBJECT);

	/* GObject signals */
	gobject_class->dispose = gwt_plot_scale_dispose;
	gobject_class->finalize = gwt_plot_scale_finalize;

	/* GtkWidget signals */
	widget_class->expose_event = gwt_plot_scale_expose;
	widget_class->button_press_event = gwt_plot_scale_button_press;
	widget_class->button_release_event = gwt_plot_scale_button_release;
	widget_class->scroll_event = gwt_plot_scale_scroll_event;
	widget_class->motion_notify_event = gwt_plot_scale_motion_notify;
	widget_class->size_request = gwt_plot_scale_size_request;
	//widget_class->size_allocate = gwt_plot_scale_size_allocate;


	gobject_class->set_property = gwt_plot_scale_set_property;
	gobject_class->get_property = gwt_plot_scale_get_property;

	g_object_class_install_property (gobject_class,
			PROP_SCALE_TYPE,
			g_param_spec_enum ("scale_type",
				"Scale Type",
				"The scale type",
				GWT_TYPE_PLOT_SCALE_TYPE, 
				GWT_PLOT_SCALE_LEFT, 
				G_PARAM_READWRITE));


#if 0
	/* GwtPlotCanvas signals */
	egg_clock_face_signals[TIME_CHANGED] = g_signal_new (
			"time-changed",
			G_OBJECT_CLASS_TYPE (obj_class),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET (EggClockFaceClass, time_changed),
			NULL, NULL,
			_clock_marshal_VOID__INT_INT,
			G_TYPE_NONE, 2,
			G_TYPE_INT,
			G_TYPE_INT);
#endif
	g_type_class_add_private (gobject_class, sizeof (GwtPlotScalePrivate));
}

static void
gwt_plot_scale_init (GwtPlotScale *scale)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	gtk_widget_add_events (GTK_WIDGET (scale),
			GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK | GDK_SCROLL_MASK);



	priv->dispose_has_run = FALSE;
	priv->tick_base = 10;
	priv->minor_ticks = 10;


	priv->gap_size = 2;
	priv->tick_size = 10;
	priv->text_size = 15;
	priv->label_size = 15;

	priv->format_callback = default_format_callback;
}


static void
gwt_plot_scale_draw_background(GtkWidget *widget, cairo_t* cr)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	/* background color */
	cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
	cairo_rectangle (cr, 0, 0, widget->allocation.width, widget->allocation.height);
	cairo_fill (cr);	
}

static gdouble
gwt_plot_scale_calc_tick_delta(GwtPlotScale* scale, gdouble min, gdouble max, gdouble mag)
{
	GwtPlotScalePrivate *priv;
	gdouble tick_delta, d;
	int i;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	tick_delta = priv->tick_base;

	/* try to find a delta that causes the device coordinates distance to be between
	 * 50 and 200 pixel. Use a loop to prefent endless jittering between just to small 
	 * and just to big */
	for (i = 0; i < 20; i++) {
		d = gwt_plot_adjustment_user_to_device_distance(priv->adjustment, tick_delta);

		if (d < min)
			tick_delta *= mag;
		else if (d > max)
			tick_delta /= mag;
		else
			break;
	}

	d = gwt_plot_adjustment_user_to_device_distance(priv->adjustment, tick_delta);

	if (d < min)
		tick_delta *= mag;

	return tick_delta;
}

static void
gwt_plot_scale_draw_xscale(GtkWidget *widget, cairo_t* cr)
{
	GwtPlotScalePrivate *priv;
	gdouble x, dx, x_max, x_min, dev_x;
	gdouble tick_start, tick_end, minor_tick_start, minor_tick_end, text_start, backbone_pos;
	guint i;
	char tmp[128];
	cairo_text_extents_t te;
	cairo_font_extents_t fe;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	g_return_if_fail (priv->adjustment != NULL);

	gwt_plot_adjustment_set_device_interval(priv->adjustment, 0, widget->allocation.width);

	x_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
	x_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

	cairo_font_extents(cr, &fe);

	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_TOP:
		tick_end = widget->allocation.height - priv->gap_size;
		minor_tick_end = widget->allocation.height - priv->gap_size;
		tick_start = tick_end - priv->tick_size;
		minor_tick_start = minor_tick_end - (priv->tick_size / 2.0);
		backbone_pos = tick_end - 0.5;
		text_start = tick_start - 2.0;
		break;

	case GWT_PLOT_SCALE_BOTTOM:
		tick_start = priv->gap_size;
		minor_tick_start = priv->gap_size;
		tick_end = tick_start + priv->tick_size;
		minor_tick_end = minor_tick_start + (priv->tick_size / 2.0);
		backbone_pos = tick_start + 0.5;
		text_start = tick_end + 2.0 + fe.height;
		break;
	default:
		return;
	};

	cairo_set_source_rgb (cr, 0, 0, 0);
	cairo_set_line_width (cr, 1.0);

	//cairo_set_antialias (cr , CAIRO_ANTIALIAS_NONE);

	cairo_move_to(cr, 0, backbone_pos);
	cairo_line_to(cr, widget->allocation.width, backbone_pos);
	cairo_stroke (cr);


	priv->format_callback(GWT_PLOT_SCALE(widget), G_MAXDOUBLE, tmp, sizeof(tmp));
	tmp[sizeof(tmp)-1] = 0;
	cairo_text_extents(cr, tmp, &te); 
	dx = gwt_plot_scale_calc_tick_delta(GWT_PLOT_SCALE(widget), te.x_advance * 1.5, widget->allocation.width / 2 , 2);

	x = (trunc(x_min / dx) * dx) - dx;
	while(x <= x_max) {
		priv->format_callback(GWT_PLOT_SCALE(widget), x, tmp, sizeof(tmp));
		tmp[sizeof(tmp)-1] = 0;

		dev_x = gwt_plot_adjustment_user_to_device(priv->adjustment, x);
		if (dev_x < widget->allocation.width / 2)
			dev_x += 0.5;
		else
			dev_x -= 0.5;	

		cairo_move_to (cr, dev_x, tick_start);
		cairo_line_to (cr, dev_x, tick_end);
		cairo_stroke(cr);

		if (dev_x >= 0 && dev_x <= widget->allocation.width) {
			cairo_text_extents(cr, tmp, &te); 
			if (dev_x - (te.x_advance / 2.0) < 0)
				dev_x = (te.x_advance / 2.0);
			if (dev_x > widget->allocation.width - (te.x_advance / 2.0))
				dev_x = widget->allocation.width - (te.x_advance / 2.0);

			cairo_move_to (cr, dev_x - (te.x_advance / 2.0), text_start);
			cairo_show_text(cr, tmp);
		}

		if (gwt_plot_adjustment_user_to_device_distance(priv->adjustment, dx) >= 50) {
			for (i = 1; i < priv->minor_ticks; i++) {
				dev_x = gwt_plot_adjustment_user_to_device(priv->adjustment, x + ((dx / (gdouble)priv->minor_ticks) * (gdouble)i));
				if (dev_x < widget->allocation.width / 2)
					dev_x += 0.5;
				else
					dev_x -= 0.5;	

				cairo_move_to (cr, dev_x, minor_tick_start);
				cairo_line_to (cr, dev_x, minor_tick_end);
				cairo_stroke(cr);
			}
		}

		x += dx;
	}

	cairo_stroke (cr);
}

static void
gwt_plot_scale_draw_yscale(GtkWidget *widget, cairo_t* cr)
{
	GwtPlotScalePrivate *priv;
	gdouble y, dy, y_max, y_min, dev_y;
	gdouble tick_start, tick_end, minor_tick_start, minor_tick_end, text_start, backbone_pos;
	gdouble max_text_size = 0.0;
	guint i;
	char tmp[128];
	cairo_text_extents_t te;
	cairo_font_extents_t fe;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	g_return_if_fail (priv->adjustment != NULL);

	gwt_plot_adjustment_set_device_interval(priv->adjustment, 0, widget->allocation.height);

	y_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
	y_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

	cairo_font_extents(cr, &fe);

	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_LEFT:
		tick_end = widget->allocation.width - priv->gap_size;
		minor_tick_end = widget->allocation.width - priv->gap_size;
		tick_start = tick_end - priv->tick_size;
		minor_tick_start = minor_tick_end - (priv->tick_size / 2.0);
		backbone_pos = tick_end - 0.5;
		break;

	case GWT_PLOT_SCALE_RIGHT:
		tick_start = priv->gap_size;
		minor_tick_start = priv->gap_size;
		tick_end = tick_start + priv->tick_size;
		minor_tick_end = minor_tick_start + (priv->tick_size / 2.0);
		backbone_pos = tick_start + 0.5;
		break;
	default:
		return;
	};

	cairo_set_source_rgb (cr, 0, 0, 0);
	cairo_set_line_width (cr, 1.0);

	cairo_move_to(cr, backbone_pos, 0);
	cairo_line_to(cr, backbone_pos, widget->allocation.height);
	cairo_stroke (cr);

	dy = gwt_plot_scale_calc_tick_delta(GWT_PLOT_SCALE(widget), fe.height * 1.5, widget->allocation.height / 2, 2);

	y = (trunc(y_min / dy) * dy) - dy; /* starting Y value rounded to dy  */
	while (y <= y_max) {
		dev_y = widget->allocation.height - gwt_plot_adjustment_user_to_device(priv->adjustment, y);
		if (dev_y < widget->allocation.height / 2)
			dev_y += 0.5;
		else
			dev_y -= 0.5;	

		cairo_move_to (cr, tick_start, dev_y);
		cairo_line_to (cr, tick_end, dev_y);
		cairo_stroke(cr);

		if (dev_y >= 0 && dev_y <= widget->allocation.height) {
			priv->format_callback(GWT_PLOT_SCALE(widget), y, tmp, sizeof(tmp));
			tmp[sizeof(tmp)-1] = 0;

			cairo_text_extents(cr, tmp, &te); 
			if (te.x_advance > max_text_size)
				max_text_size = te.x_advance;

			switch (priv->scale_type) {
			case GWT_PLOT_SCALE_LEFT:
				text_start = tick_start - te.x_advance - 2;
				break;
			case GWT_PLOT_SCALE_RIGHT:
				text_start = tick_end + 2;
				break;
			default:
				return;
			}
			
			if (dev_y - (te.height / 2.0) < 0)
				dev_y = (te.height / 2.0);
			if (dev_y > widget->allocation.height - (te.height / 2.0))
				dev_y = widget->allocation.height - (te.height / 2.0);

			cairo_move_to (cr, text_start, dev_y + (te.height / 2.0));
			cairo_show_text(cr, tmp);

		}

		if (gwt_plot_adjustment_user_to_device_distance(priv->adjustment, dy) >= 50) {
			for (i = 1; i < priv->minor_ticks; i++) {
				dev_y = widget->allocation.height -  gwt_plot_adjustment_user_to_device(priv->adjustment, y + ((dy / (gdouble)priv->minor_ticks) * (gdouble)i));
				if (dev_y < widget->allocation.height / 2)
					dev_y += 0.5;
				else
					dev_y -= 0.5;	

				cairo_move_to (cr, minor_tick_start, dev_y);
				cairo_line_to (cr, minor_tick_end, dev_y);
				cairo_stroke(cr);
			}
		}

		y += dy;
	}

	priv->text_size = max_text_size + 4;
}

static void
gwt_plot_scale_draw_scale(GtkWidget *widget, cairo_t* cr)
{
	GwtPlotScalePrivate *priv;
	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	g_return_if_fail (priv->adjustment != NULL);

	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_TOP:
	case GWT_PLOT_SCALE_BOTTOM:
		gwt_plot_scale_draw_xscale(widget, cr);		
		break;

	case GWT_PLOT_SCALE_LEFT:
	case GWT_PLOT_SCALE_RIGHT:
		gwt_plot_scale_draw_yscale(widget, cr);
		break;
	default:
		return;
	};
}

static void
gwt_plot_scale_draw_label(GtkWidget *widget, cairo_t* cr)
{
	GwtPlotScalePrivate *priv;
	gdouble xpos = 0, ypos = 0;
	cairo_text_extents_t te;
	cairo_font_extents_t fe;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	g_return_if_fail (priv->adjustment != NULL);

	if (priv->label == NULL)
		return;

	cairo_set_source_rgb (cr, priv->label_color[0], priv->label_color[1], priv->label_color[2]);
	//cairo_select_font_face (cr, "Georgia", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
	cairo_set_font_size (cr, 14);
	cairo_text_extents (cr, priv->label, &te);
	cairo_font_extents(cr, &fe);


	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_TOP:
		xpos = (widget->allocation.width / 2.0) - (te.width / 2.0);
		ypos = te.height;	
		cairo_move_to (cr, xpos, ypos);
		priv->label_size = fe.height;
		break;

	case GWT_PLOT_SCALE_BOTTOM:
		xpos = (widget->allocation.width / 2.0) - (te.width / 2.0);
		ypos = priv->gap_size + priv->tick_size + priv->text_size + te.height;	
		cairo_move_to (cr, xpos, ypos);
		priv->label_size = fe.height;
		break;

	case GWT_PLOT_SCALE_LEFT:
		xpos = fe.height;//widget->allocation.width - (priv->gap_size  + priv->tick_size + priv->text_size);
		ypos = (widget->allocation.height / 2.0) + (te.width / 2.0);
		cairo_move_to (cr, xpos, ypos);
		cairo_rotate(cr, (M_PI / 2) + M_PI);
		priv->label_size = fe.height;
		break;

	case GWT_PLOT_SCALE_RIGHT:
		xpos = widget->allocation.width - fe.height; //priv->gap_size + priv->tick_size + priv->text_size;	
		ypos = (widget->allocation.height / 2.0) - (te.width / 2.0);
		cairo_move_to (cr, xpos, ypos);
		cairo_rotate(cr, (M_PI / 2));
		priv->label_size = fe.height;
		break;
	};

	cairo_show_text(cr , priv->label);
}

static gboolean
gwt_plot_scale_expose (GtkWidget *scale, GdkEventExpose *event)
{
	cairo_t *cr;
	GwtPlotScalePrivate *priv;
	gdouble current_size;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	/* get a cairo_t */
	cr = gdk_cairo_create (scale->window);

	cairo_rectangle (cr,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
	cairo_clip (cr);
	
	gwt_plot_scale_draw_background(scale, cr);

	/* if no adjustment is set, only draw the background */
	if (priv->adjustment == NULL)
		return TRUE;

	gwt_plot_scale_draw_scale(scale, cr);
	
	gwt_plot_scale_draw_label(scale, cr);

	cairo_destroy (cr);

	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_TOP:
	case GWT_PLOT_SCALE_BOTTOM:
		current_size = scale->allocation.height;
		break;
	case GWT_PLOT_SCALE_LEFT:
	case GWT_PLOT_SCALE_RIGHT:
		current_size = scale->allocation.width;
		break;
	default:
		return TRUE;
	}
	
	if (priv->gap_size + priv->tick_size + priv->text_size + priv->label_size > current_size)
		gtk_widget_queue_resize(scale);

	return FALSE;
}

static gboolean
gwt_plot_scale_scroll_event(GtkWidget *scale, GdkEventScroll *event) 
{
	GwtPlotScalePrivate *priv;
	gdouble d, x_min, x_max, y_min, y_max;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);
	
	switch (event->direction) {
	case GDK_SCROLL_DOWN:
	case GDK_SCROLL_LEFT:
		d = 1;
		break;
	case GDK_SCROLL_UP:
	case GDK_SCROLL_RIGHT:
		d = -1;
		break;
	};



	switch (priv->scale_type) {
	case GWT_PLOT_SCALE_TOP:
	case GWT_PLOT_SCALE_BOTTOM:
		x_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
		x_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

		d = d * ((gdouble)scale->allocation.width / 4.0);
		d = d / (gdouble)scale->allocation.width;
		d = (x_max - x_min) * d;

		gwt_plot_adjustment_move(priv->adjustment, -d);			
		break;

	case GWT_PLOT_SCALE_LEFT:
	case GWT_PLOT_SCALE_RIGHT:
		y_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
		y_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

		d = d * ((gdouble)scale->allocation.height / 4.0);
		d = d / (gdouble)scale->allocation.height;
		d = (y_max - y_min) * d;

		gwt_plot_adjustment_move(priv->adjustment, d);			
		break;
	};

	return FALSE;
}


static gboolean
gwt_plot_scale_button_press (GtkWidget *scale, GdkEventButton *event)
{
	GwtPlotScalePrivate *priv;
	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	priv->drag_start_x = event->x;
	priv->drag_start_y = event->y;
	priv->dragging = TRUE;


	return FALSE;
}

static void
gwt_plot_scale_redraw_canvas (GwtPlotScale *scale)
{
	GtkWidget *widget;
	GdkRegion *region;
	
	widget = GTK_WIDGET (scale);

	if (!widget->window) return;

	region = gdk_drawable_get_clip_region (widget->window);
	/* redraw the cairo canvas completely by exposing it */
	gdk_window_invalidate_region (widget->window, region, TRUE);
	gdk_window_process_updates (widget->window, TRUE);

	gdk_region_destroy (region);
}

static gboolean
gwt_plot_scale_motion_notify (GtkWidget *scale, GdkEventMotion *event)
{
	GwtPlotScalePrivate *priv;
	gdouble d, x_min, x_max, y_min, y_max;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	if (priv->dragging)
	{
		switch (priv->scale_type) {
		case GWT_PLOT_SCALE_TOP:
		case GWT_PLOT_SCALE_BOTTOM:
			x_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
			x_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

			d = (event->x - priv->drag_start_x) / (gdouble)scale->allocation.width;
			d = (x_max - x_min) * d;

			gwt_plot_adjustment_move(priv->adjustment, -d);
			break;

		case GWT_PLOT_SCALE_LEFT:
		case GWT_PLOT_SCALE_RIGHT:
			y_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
			y_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

			d = (event->y - priv->drag_start_y) / (gdouble)scale->allocation.height;
			d = (y_max - y_min) * d;

			gwt_plot_adjustment_move(priv->adjustment, d);
			break;
		};

		priv->drag_start_x = event->x;
		priv->drag_start_y = event->y;

	}

	return FALSE;
}

static gboolean
gwt_plot_scale_button_release (GtkWidget *scale, GdkEventButton *event)
{
	GwtPlotScalePrivate *priv;
	gdouble d, x_min, x_max, y_min, y_max;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);
	
	if (priv->dragging)
	{
		priv->dragging = FALSE;

		switch (priv->scale_type) {
		case GWT_PLOT_SCALE_TOP:
		case GWT_PLOT_SCALE_BOTTOM:
			x_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
			x_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

			d = (event->x - priv->drag_start_x) / (gdouble)scale->allocation.width;
			d = (x_max - x_min) * d;

			gwt_plot_adjustment_move(priv->adjustment, -d);			
			break;

		case GWT_PLOT_SCALE_LEFT:
		case GWT_PLOT_SCALE_RIGHT:
			y_min = gwt_plot_adjustment_get_start_value(priv->adjustment);
			y_max = gwt_plot_adjustment_get_end_value(priv->adjustment);

			d = (event->y - priv->drag_start_y) / (gdouble)scale->allocation.height;
			d = (y_max - y_min) * d;

			gwt_plot_adjustment_move(priv->adjustment, d);			
			break;
		};
	}

	return FALSE;
}

static gboolean
gwt_plot_scale_update (gpointer data)
{
	GwtPlotScale *scale;
	GwtPlotScalePrivate *priv;

	scale = GWT_PLOT_SCALE (data);
	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);
		
	gwt_plot_scale_redraw_canvas (scale);

	return TRUE; /* keep running this event */
}

static void
gwt_plot_scale_size_request (GtkWidget      *widget,
			GtkRequisition *requisition)
{
	GwtPlotScale *scale;
	GwtPlotScalePrivate *priv;
  
	g_return_if_fail (GWT_IS_PLOT_SCALE (widget));
	g_return_if_fail (requisition != NULL);
  
	scale = GWT_PLOT_SCALE (widget);
	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	if (priv->adjustment == NULL) {
		requisition->width = 0;
		requisition->height = 0;
	} else {
		switch (priv->scale_type) {
		case GWT_PLOT_SCALE_TOP:
		case GWT_PLOT_SCALE_BOTTOM:
			requisition->width = 1;	
			requisition->height = priv->gap_size + priv->tick_size + priv->text_size + priv->label_size;
			break;

		case GWT_PLOT_SCALE_LEFT:
		case GWT_PLOT_SCALE_RIGHT:
			requisition->width = priv->gap_size + priv->tick_size + priv->text_size + priv->label_size;
			requisition->height = 1;
			break;
		}
	}
}

static void
gwt_plot_scale_size_allocate (GtkWidget     *widget,
                         GtkAllocation *allocation)
{
	GwtPlotScale *scale;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GWT_IS_PLOT_SCALE (widget));
	g_return_if_fail (allocation != NULL);


	scale = GWT_PLOT_SCALE (widget);

	//(* GTK_WIDGET_CLASS (gwt_plot_canvas_parent_class)->size_allocate) (widget, allocation);


	widget->allocation = *allocation;
	if (GTK_WIDGET_REALIZED (widget)) {
		gdk_window_move_resize (widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);
	}
}

static void
gwt_plot_scale_adjustment_changed_cb(GwtPlotAdjustment* adj, GwtPlotScale* scale)
{
	g_return_if_fail (adj != NULL);
	g_return_if_fail (GWT_IS_PLOT_ADJUSTMENT (adj));
	g_return_if_fail (scale != NULL);
	g_return_if_fail (GWT_IS_PLOT_SCALE (scale));

	//g_debug("gwt_plot_scale_adjustment_changed_cb %p %p", adj, scale);

	gtk_widget_queue_draw(GTK_WIDGET(scale));
}


GtkWidget *
gwt_plot_scale_new (void)
{
	return g_object_new (GWT_TYPE_PLOT_SCALE, NULL);
}


void
gwt_plot_scale_set_type(GwtPlotScale* scale, GwtPlotScaleType type)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	priv->scale_type = type;
}

void 
gwt_plot_scale_set_adjustment (GwtPlotScale* scale, GwtPlotAdjustment* adj)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	if (priv->adjustment != NULL) {
		g_signal_handlers_disconnect_by_func(priv->adjustment, G_CALLBACK (gwt_plot_scale_adjustment_changed_cb), scale);

		g_object_unref (priv->adjustment);
	}

	priv->adjustment = adj;	

	if (adj != NULL) {
		g_object_ref_sink (adj);
		g_signal_connect (priv->adjustment, "changed", G_CALLBACK (gwt_plot_scale_adjustment_changed_cb), scale);
	}
}

GwtPlotAdjustment* 
gwt_plot_scale_get_adjustment(GwtPlotScale* scale)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	return priv->adjustment;
}

static void
gwt_plot_scale_get_property (GObject *object, guint prop_id, GValue *value, 
	                     GParamSpec *pspec)
{
 	GwtPlotScalePrivate *priv;
	GwtPlotScale *scale = GWT_PLOT_SCALE (object);

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	switch (prop_id)
	{
	case PROP_SCALE_TYPE:
		g_value_set_enum (value, priv->scale_type);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gwt_plot_scale_set_property (GObject * object, guint prop_id, 
	                     const GValue * value, GParamSpec * pspec)
{
 	GwtPlotScalePrivate *priv;
	GwtPlotScale *scale = GWT_PLOT_SCALE (object);

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	switch (prop_id) 
	{	
	case PROP_SCALE_TYPE:
		priv->scale_type = g_value_get_enum (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}


void
gwt_plot_scale_set_label(GwtPlotScale* scale, const char* label) 
{
	GwtPlotScalePrivate *priv;
	int len;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	if (priv->label != NULL) {
		g_free(priv->label);
		priv->label = NULL;
	}

	if (label != NULL) {
		len = strlen(label);
		priv->label = g_malloc(len + 1);
		strcpy(priv->label, label);	
	}

	gtk_widget_queue_draw(GTK_WIDGET(scale));
}

void gwt_plot_scale_set_label_rgb(GwtPlotScale* widget, gdouble r, gdouble g, gdouble b)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (widget);

	priv->label_color[0]= r;
	priv->label_color[1] = g;
	priv->label_color[2] = b;

	gtk_widget_queue_draw(GTK_WIDGET(widget));
}

void gwt_plot_scale_set_format_callback(GwtPlotScale* scale, GwtFormatCallback cb)
{
	GwtPlotScalePrivate *priv;

	priv = GWT_PLOT_SCALE_GET_PRIVATE (scale);

	priv->format_callback = cb;
	if (priv->format_callback == NULL)
		priv->format_callback = default_format_callback;		
}

