#include <gtk/gtk.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#include <gwt/gwt.h>

#include "can_plot_dialog.h"

#define CAN_PLOT_DIALOG_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), TYPE_CAN_PLOT_DIALOG, CanPlotDialogPrivate))

G_DEFINE_TYPE (CanPlotDialog, can_plot_dialog, GTK_TYPE_WINDOW);

static gpointer    parent_class = NULL;

typedef struct _CanPlotDialogPrivate CanPlotDialogPrivate;

#define NR_PLOTS 4
#define NR_SAMPLES 10000

/*
Temperatur von -40 bis +85 Grad
Spannung von -12 bis +12 Volt
Druck von 0 bis 50 bar
Lage von -180 bis +180 Grad
*/

const char* default_y_labels[NR_PLOTS] = {
	"Temperatur [C]",
	"Spannung [V]",
	"Druck [B]",
	"Lage [Grad]", 	

};

GdkColor default_color[NR_PLOTS] = {
	{ 0, 65535, 0, 0 },
	{ 0, 65535, 0, 65535 },
	{ 0, 0, 65535, 65535 },
	{ 0, 0, 0, 65535 },
};

struct _CanPlotDialogPrivate
{
	GtkWidget *vbox;
	GtkWidget *notebook;

	GtkWidget *x_axis_page;
	GtkWidget *x_axis_page_label;
	GtkWidget *radio_time;
	GtkWidget *radio_distance;
	GtkWidget *edit_time;
	GtkWidget *edit_distance;



	GtkWidget *y_axis_page[NR_PLOTS];
	GtkWidget *y_axis_page_label[NR_PLOTS];
	GtkWidget *edit_label[NR_PLOTS];
	GtkWidget *color_button[NR_PLOTS];
	GtkWidget *plot_type[NR_PLOTS];
	GtkWidget *line_type[NR_PLOTS];
	GtkWidget *marker_type[NR_PLOTS];


	/* plot frame */
	GtkWidget* frame;
	GtkWidget* table;
	GtkWidget* top_label;
	GtkWidget* canvas;
	GtkWidget* y_scale[NR_PLOTS];
	GtkWidget* x_scale_t;
	GtkWidget* x_scale_d;

	GtkObject* y_adjustment[NR_PLOTS];
	GtkObject* x_adjustment_t;
	GtkObject* x_adjustment_d;

	GtkWidget *event_box;

	GtkObject* plot_data_t[NR_PLOTS];
	GtkObject* plot_curve_t[NR_PLOTS];
	GtkObject* plot_grid_t[NR_PLOTS];

	GtkObject* plot_data_d[NR_PLOTS];
	GtkObject* plot_curve_d[NR_PLOTS];
	GtkObject* plot_grid_d[NR_PLOTS];

	gdouble plot_color_red[NR_PLOTS];
	gdouble plot_color_green[NR_PLOTS];
	gdouble plot_color_blue[NR_PLOTS];
};

enum
{
	TIME_CHANGED,
	LAST_SIGNAL
};

//static guint gwt_plot_canvas_signals[LAST_SIGNAL] = { 0 };


static gint format_time_cb( GwtPlotScale* scale, gdouble value, char* buffer, guint buffer_size) 
{
	gdouble h,m,s;
	gint sign = 0;
	gint res;



	if (value == G_MAXDOUBLE) {
		h = 999;
		m = 59;
		s = 59.9999;
	} else {
		if (value < 0.0)	
			sign = -1;
		else
			sign = 1;

		value = fabs(value);
		h = floor(value / 3600.0);	
		m = floor((value - h * 60) / 60);
		s = value - h * 3600 - m * 60;

		if (sign < 0)
			h = -h;
	}


	if ( h == 0) {
		if (m == 0) {
			if (sign < 0)
				s  = -s;
			res = snprintf(buffer, buffer_size, "%gs", s);	
		} else {
			if (sign < 0)
				m = -m;

			res = snprintf(buffer, buffer_size, "%2dm%gs", (guint)m, s);	
		}
 	} else {
		if (sign < 0)
			h = -h;
		
		res = snprintf(buffer, buffer_size, "%3dh%2dm%gs", (guint)h, (guint)m, s);	
	}

	return res;
}

static gint format_distance_cb( GwtPlotScale* scale, gdouble value, char* buffer, guint buffer_size) 
{
	if (value == G_MAXDOUBLE)
		value = 99999.9;

	return snprintf(buffer, buffer_size, "%gm", value);
}

static void create_plots(CanPlotDialog *dlg) {
	CanPlotDialogPrivate *priv;
	int i,k;
	gdouble t, d, v, r, g, b;
	GwtPlotDataPoint data_point;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	priv->x_adjustment_t = gwt_plot_adjustment_new(0, 500);
	priv->x_adjustment_d = gwt_plot_adjustment_new(0, 500);


	for (i = 0; i < NR_PLOTS; i++) {
		priv->y_adjustment[i] = gwt_plot_adjustment_new(-100, 100);

		priv->plot_curve_t[i] = gwt_plot_curve_new();
		priv->plot_data_t[i] = gwt_plot_data_new();
		priv->plot_grid_t[i] = gwt_plot_grid_new();

		priv->plot_curve_d[i] = gwt_plot_curve_new();
		priv->plot_data_d[i] = gwt_plot_data_new();
		priv->plot_grid_d[i] = gwt_plot_grid_new();


		priv->plot_color_red[i] = (i & 0x1) ? 0.8 : 0.2;
		priv->plot_color_green[i] = (i >> 1) & 0x1 ? 0.8 : 0.2;
		priv->plot_color_blue[i] = (i >> 2) & 0x1 ? 0.8 : 0.2;

		r = priv->plot_color_red[i];
		g = priv->plot_color_green[i];
		b = priv->plot_color_blue[i];

		gwt_plot_item_set_rgba (GWT_PLOT_ITEM(priv->plot_curve_d[i]), r, g, b, 1.0);
		gwt_plot_item_set_rgba (GWT_PLOT_ITEM(priv->plot_curve_t[i]), r, g, b, 1.0);

		gwt_plot_item_set_rgba (GWT_PLOT_ITEM(priv->plot_grid_d[i]), r, g, b, 1.0);
		gwt_plot_item_set_rgba (GWT_PLOT_ITEM(priv->plot_grid_t[i]), r, g, b, 1.0);

		gwt_plot_curve_set_symbol( GWT_PLOT_CURVE(priv->plot_curve_d[i]), GWT_PLOT_CURVE_SYMBOL_SQUARE);
		gwt_plot_curve_set_symbol( GWT_PLOT_CURVE(priv->plot_curve_t[i]), GWT_PLOT_CURVE_SYMBOL_DIAMOND);

		gwt_plot_curve_set_style( GWT_PLOT_CURVE(priv->plot_curve_d[i]), GWT_PLOT_CURVE_STYLE_STICKS);
		gwt_plot_curve_set_style( GWT_PLOT_CURVE(priv->plot_curve_t[i]), GWT_PLOT_CURVE_STYLE_FITTED);

		gwt_plot_curve_set_line_style( GWT_PLOT_CURVE(priv->plot_curve_d[i]), GWT_PLOT_CURVE_LINE_STYLE_DOT);
		gwt_plot_curve_set_line_style( GWT_PLOT_CURVE(priv->plot_curve_t[i]), GWT_PLOT_CURVE_LINE_STYLE_DOT_DASH);
	}


	t = -11.0;
	d = -11.0;
	for (k = 0; k < NR_SAMPLES; k++) {
		/* Temperatur von -40 bis +85 Grad */
		v = ((gdouble)(rand() % 1000) + 5000) / 100.0;  
		data_point.x = t;
		data_point.y = v;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_t[0]), &data_point);

		data_point.x = d;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_d[0]), &data_point);

		/* Spannung von -12 bis +12 Volt */
		v = ((gdouble)(rand() % 100) + 1000 ) / 100.0;  
		data_point.x = t;
		data_point.y = v;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_t[1]), &data_point);

		data_point.x = d;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_d[1]), &data_point);


		/* Druck von 0 bis 50 bar */
		v = ((gdouble)(rand() % 3000) + 1000) / 100.0;  
		data_point.x = t;
		data_point.y = v;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_t[2]), &data_point);

		data_point.x = d;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_d[2]), &data_point);


		/* Lage von -180 bis +180 Grad */
		v = ((gdouble)(rand() % 200) - 100) / 10.0;  
		data_point.x = t;
		data_point.y = v;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_t[3]), &data_point);

		data_point.x = d;
		gwt_plot_data_insert( GWT_PLOT_DATA(priv->plot_data_d[3]), &data_point);

	


		t += 0.1;	// 100 ms
		d += 0.01 + ((rand() % 999) + 1) / 100000.0;
	}

	for (i = 0; i < NR_PLOTS; i++) {
		gwt_plot_curve_set_data (GWT_PLOT_CURVE(priv->plot_curve_t[i]), GWT_PLOT_DATA(priv->plot_data_t[i]));
		gwt_plot_item_set_x_adjustment(GWT_PLOT_ITEM(priv->plot_curve_t[i]), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t));	
		gwt_plot_item_set_y_adjustment(GWT_PLOT_ITEM(priv->plot_curve_t[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));

		gwt_plot_item_set_x_adjustment(GWT_PLOT_ITEM(priv->plot_grid_t[i]), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t));	
		gwt_plot_item_set_y_adjustment(GWT_PLOT_ITEM(priv->plot_grid_t[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));

		gwt_plot_curve_set_data (GWT_PLOT_CURVE(priv->plot_curve_d[i]), GWT_PLOT_DATA(priv->plot_data_d[i]));
		gwt_plot_item_set_x_adjustment(GWT_PLOT_ITEM(priv->plot_curve_d[i]), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d));
		gwt_plot_item_set_y_adjustment(GWT_PLOT_ITEM(priv->plot_curve_d[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));
		gwt_plot_item_set_x_adjustment(GWT_PLOT_ITEM(priv->plot_grid_d[i]), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d));
		gwt_plot_item_set_y_adjustment(GWT_PLOT_ITEM(priv->plot_grid_d[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));


		//gwt_plot_data_find_min_max_y( GWT_PLOT_DATA(priv->plot_data_d[i]), min_x, max_x, &min_y, &max_y);
	}

	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[0]), -50); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[0]), 100); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[0]), -50); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[0]), 100); 

	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[1]), -20); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[1]), 20); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[1]), -20); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[1]), 20); 

	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[2]), 0); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[2]), 100); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[2]), 0); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[2]), 100); 

	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[3]), -200); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[3]), 200); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[3]), -200); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->y_adjustment[3]), 200); 




	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t), -10); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t), t); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t), -10); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t), 10); 

	gwt_plot_adjustment_set_min_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d), -10); 
	gwt_plot_adjustment_set_max_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d), d); 
	gwt_plot_adjustment_set_start_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d), -10); 
	gwt_plot_adjustment_set_end_value(GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d), -5); 
}


static void
can_plot_dialog_class_init (CanPlotDialogClass *class)
{
	GObjectClass *gobject_class;
	GtkWidgetClass *widget_class;

	gobject_class = G_OBJECT_CLASS (class);
	widget_class = GTK_WIDGET_CLASS (class);

	parent_class = g_type_class_ref (G_TYPE_OBJECT);

	/* GObject signals */

	/* GtkWidget signals */

#if 0
	/* GwtPlotCanvas signals */
	egg_clock_face_signals[TIME_CHANGED] = g_signal_new (
			"time-changed",
			G_OBJECT_CLASS_TYPE (obj_class),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET (EggClockFaceClass, time_changed),
			NULL, NULL,
			_clock_marshal_VOID__INT_INT,
			G_TYPE_NONE, 2,
			G_TYPE_INT,
			G_TYPE_INT);
#endif
	g_type_class_add_private (gobject_class, sizeof (CanPlotDialogPrivate));
}


void create_plot_frame(CanPlotDialog* dlg)
{
	CanPlotDialogPrivate *priv;
	int i;
	GdkColor c;
	char tmp[128];

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	priv->frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type (GTK_FRAME(priv->frame), GTK_SHADOW_IN);
	priv->event_box = gtk_event_box_new();
	c.red = 0xffff;
	c.green = 0xffff;
	c.blue = 0xffff;
	gdk_rgb_find_color(gtk_widget_get_colormap(GTK_WIDGET(priv->event_box)), &c);
	gtk_widget_modify_bg (GTK_WIDGET(priv->event_box), GTK_STATE_NORMAL, &c);
	gtk_widget_modify_bg (GTK_WIDGET(priv->event_box), GTK_STATE_ACTIVE, &c);
	gtk_widget_modify_bg (GTK_WIDGET(priv->event_box), GTK_STATE_PRELIGHT, &c);
	gtk_widget_modify_bg (GTK_WIDGET(priv->event_box), GTK_STATE_SELECTED, &c);

	priv->table = gtk_table_new (9, 4, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(priv->table), 2);

	priv->top_label = gtk_label_new("CAN Plot");
	
	priv->canvas = gwt_plot_canvas_new ();

	for (i = 0; i < NR_PLOTS; i++) {
		gwt_plot_canvas_add_item (GWT_PLOT_CANVAS(priv->canvas), GWT_PLOT_ITEM(priv->plot_curve_t[i]));
		gwt_plot_canvas_add_item (GWT_PLOT_CANVAS(priv->canvas), GWT_PLOT_ITEM(priv->plot_curve_d[i]));
		gwt_plot_canvas_add_item (GWT_PLOT_CANVAS(priv->canvas), GWT_PLOT_ITEM(priv->plot_grid_t[i]));
		gwt_plot_canvas_add_item (GWT_PLOT_CANVAS(priv->canvas), GWT_PLOT_ITEM(priv->plot_grid_d[i]));
	}


	priv->x_scale_t = gwt_plot_scale_new ();
	gwt_plot_scale_set_type(GWT_PLOT_SCALE(priv->x_scale_t), GWT_PLOT_SCALE_BOTTOM);
	sprintf(tmp, "time [sec]");
	gwt_plot_scale_set_label( GWT_PLOT_SCALE(priv->x_scale_t), tmp);
	gwt_plot_scale_set_format_callback( GWT_PLOT_SCALE(priv->x_scale_t), format_time_cb );


	priv->x_scale_d = gwt_plot_scale_new ();
	gwt_plot_scale_set_type(GWT_PLOT_SCALE(priv->x_scale_d), GWT_PLOT_SCALE_BOTTOM);
	sprintf(tmp, "distance [m]");
	gwt_plot_scale_set_label( GWT_PLOT_SCALE(priv->x_scale_d), tmp);
	gwt_plot_scale_set_format_callback( GWT_PLOT_SCALE(priv->x_scale_d), format_distance_cb );


	for (i = 0; i < NR_PLOTS;i++) {

		priv->y_scale[i] = gwt_plot_scale_new ();

		sprintf(tmp, "value%d [%c]", i+1, 'A' + i);
		gwt_plot_scale_set_label( GWT_PLOT_SCALE(priv->y_scale[i]), tmp);
		gwt_plot_scale_set_label_rgb(GWT_PLOT_SCALE(priv->y_scale[i]), 
					priv->plot_color_red[i],priv->plot_color_green[i], priv->plot_color_blue[i] );

		if (i < NR_PLOTS/2) {
			gwt_plot_scale_set_type(GWT_PLOT_SCALE(priv->y_scale[i]), GWT_PLOT_SCALE_LEFT);

			gtk_table_attach(GTK_TABLE (priv->table), priv->y_scale[i], i, i+1, 1, 2,
				GTK_FILL | GTK_SHRINK ,
				GTK_FILL | GTK_SHRINK ,
				0, 0);
		} else {
			gwt_plot_scale_set_type(GWT_PLOT_SCALE(priv->y_scale[i]), GWT_PLOT_SCALE_RIGHT);

			gtk_table_attach(GTK_TABLE (priv->table), priv->y_scale[i], i+1, i+2, 1, 2,
				GTK_FILL | GTK_SHRINK ,
				GTK_FILL | GTK_SHRINK ,
				0, 0);
		}

		gwt_plot_scale_set_adjustment(GWT_PLOT_SCALE(priv->y_scale[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));
		gwt_plot_scale_set_adjustment(GWT_PLOT_SCALE(priv->y_scale[i]), GWT_PLOT_ADJUSTMENT(priv->y_adjustment[i]));
	}

	gwt_plot_scale_set_adjustment(GWT_PLOT_SCALE(priv->x_scale_d), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_d));
	gwt_plot_scale_set_adjustment(GWT_PLOT_SCALE(priv->x_scale_t), GWT_PLOT_ADJUSTMENT(priv->x_adjustment_t));
	

	gtk_table_attach(GTK_TABLE (priv->table), priv->top_label, NR_PLOTS/2, (NR_PLOTS/2)+1, 0, 1,
			GTK_FILL | GTK_SHRINK,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->table), priv->canvas, NR_PLOTS/2, (NR_PLOTS/2)+1, 1, 2,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->table), priv->x_scale_t, NR_PLOTS/2, (NR_PLOTS/2)+1, 2, 3,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->table), priv->x_scale_d, NR_PLOTS/2, (NR_PLOTS/2)+1, 3, 4,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_container_add (GTK_CONTAINER (priv->event_box), priv->table);

	gtk_container_add (GTK_CONTAINER (priv->frame), priv->event_box);
}


void x_axis_radio_toggle_cb (GtkToggleButton *togglebutton, gpointer user_data)
{
	CanPlotDialogPrivate *priv;
	int i;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);


	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(priv->radio_time))) {
		gtk_widget_hide(GTK_WIDGET(priv->x_scale_d));
		gtk_widget_show(GTK_WIDGET(priv->x_scale_t));
		for (i=0; i < NR_PLOTS; i++) {
			gwt_plot_item_hide(GWT_PLOT_ITEM(priv->plot_curve_d[i]));
			gwt_plot_item_show(GWT_PLOT_ITEM(priv->plot_curve_t[i]));

			gwt_plot_item_hide(GWT_PLOT_ITEM(priv->plot_grid_d[i]));
			gwt_plot_item_show(GWT_PLOT_ITEM(priv->plot_grid_t[i]));
		}

	} else {
		gtk_widget_show(GTK_WIDGET(priv->x_scale_d));
		gtk_widget_hide(GTK_WIDGET(priv->x_scale_t));

		for (i=0; i < NR_PLOTS; i++) {
			gwt_plot_item_show(GWT_PLOT_ITEM(priv->plot_curve_d[i]));
			gwt_plot_item_hide(GWT_PLOT_ITEM(priv->plot_curve_t[i]));
			gwt_plot_item_show(GWT_PLOT_ITEM(priv->plot_grid_d[i]));
			gwt_plot_item_hide(GWT_PLOT_ITEM(priv->plot_grid_t[i]));
		}
	}
}

void x_axis_label_edit_cb(GParamSpec* spec, GObject* obj, gpointer user_data) {
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	gwt_plot_scale_set_label(GWT_PLOT_SCALE(priv->x_scale_t), gtk_entry_get_text(GTK_ENTRY(priv->edit_time)));
	gwt_plot_scale_set_label(GWT_PLOT_SCALE(priv->x_scale_d), gtk_entry_get_text(GTK_ENTRY(priv->edit_distance)));
}

void
create_x_axis_page(CanPlotDialog *dlg)
{
	CanPlotDialogPrivate *priv;
	GtkWidget *alignment;

	GtkWidget *label;
	GSList* group = NULL;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	alignment = gtk_alignment_new (0, 0, 0, 0);

	priv->x_axis_page =  gtk_table_new(2, 2, FALSE);

	gtk_container_add (GTK_CONTAINER(alignment), priv->x_axis_page);


	priv->x_axis_page_label = gtk_label_new("X-Axis Settings");


	priv->radio_time = gtk_radio_button_new_with_label(group, "Use Time Scale");
	group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (priv->radio_time));	
	label = gtk_label_new("Label:");
	priv->edit_time = gtk_entry_new();


	gtk_table_attach(GTK_TABLE (priv->x_axis_page), priv->radio_time, 0, 2, 0, 1,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->x_axis_page), label, 0, 1, 1, 2,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->x_axis_page), priv->edit_time, 1, 2, 1, 2,
			GTK_SHRINK | GTK_EXPAND,
			GTK_SHRINK,
			0, 0);


	priv->radio_distance = gtk_radio_button_new_with_label(group, "Use Distance Scale");
	label = gtk_label_new("Label:");
	priv->edit_distance = gtk_entry_new();

	gtk_table_attach(GTK_TABLE (priv->x_axis_page), priv->radio_distance, 2, 4, 0, 1,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->x_axis_page), label, 2, 3, 1, 2,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->x_axis_page), priv->edit_distance, 3, 4, 1, 2,
			GTK_SHRINK | GTK_EXPAND,
			GTK_SHRINK,
			0, 0);


	gtk_notebook_append_page( GTK_NOTEBOOK(priv->notebook), alignment, priv->x_axis_page_label);



	g_signal_connect (G_OBJECT (priv->radio_time), "toggled",
		      G_CALLBACK (x_axis_radio_toggle_cb), (gpointer) dlg);

//	g_signal_connect (G_OBJECT (priv->radio_distance), "toggled",
//		      G_CALLBACK (x_axis_radio_toggle_cb), (gpointer) dlg);

	g_signal_connect (G_OBJECT (priv->edit_distance), "notify::text", G_CALLBACK (x_axis_label_edit_cb), dlg);

	g_signal_connect (G_OBJECT (priv->edit_time), "notify::text", G_CALLBACK (x_axis_label_edit_cb), dlg);
}


void y_axis_label_edit_cb(GParamSpec* spec, GObject* obj, gpointer user_data) {
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	for (i=0; i < NR_PLOTS; i++) {
		gwt_plot_scale_set_label(GWT_PLOT_SCALE(priv->y_scale[i]), gtk_entry_get_text(GTK_ENTRY(priv->edit_label[i])));

		gtk_label_set_text(GTK_LABEL(priv->y_axis_page_label[i]), gtk_entry_get_text(GTK_ENTRY(priv->edit_label[i])));
	}
}

void color_button_cb(GtkColorButton* button, gpointer user_data) {
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	for (i=0; i < NR_PLOTS; i++) {
		GdkColor color;
		gtk_color_button_get_color(GTK_COLOR_BUTTON(priv->color_button[i]), &color);

		gwt_plot_item_set_rgba(GWT_PLOT_ITEM(priv->plot_curve_t[i]), (double)color.red / 65535.0, (double)color.green / 65535.0, (double)color.blue / 65535.0, 1.0);
		gwt_plot_item_set_rgba(GWT_PLOT_ITEM(priv->plot_grid_t[i]), (double)color.red / 65535.0, (double)color.green / 65535.0, (double)color.blue / 65535.0, 1.0);
		gwt_plot_item_set_rgba(GWT_PLOT_ITEM(priv->plot_curve_d[i]), (double)color.red / 65535.0, (double)color.green / 65535.0, (double)color.blue / 65535.0, 1.0);
		gwt_plot_item_set_rgba(GWT_PLOT_ITEM(priv->plot_grid_d[i]), (double)color.red / 65535.0, (double)color.green / 65535.0, (double)color.blue / 65535.0, 1.0);
		gwt_plot_scale_set_label_rgb(GWT_PLOT_SCALE(priv->y_scale[i]), (double)color.red / 65535.0, (double)color.green / 65535.0, (double)color.blue / 65535.0);
	}
}

void plot_style_combo_cb(GtkComboBox *combo, gpointer user_data)
{
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	for (i=0; i < NR_PLOTS; i++) {
		int index = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->plot_type[i]));

		gwt_plot_curve_set_style(GWT_PLOT_CURVE(priv->plot_curve_t[i]), index);
		gwt_plot_curve_set_style(GWT_PLOT_CURVE(priv->plot_curve_d[i]), index);
	}
}

void plot_line_style_combo_cb(GtkComboBox *combo, gpointer user_data)
{
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	for (i=0; i < NR_PLOTS; i++) {
		int index = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->line_type[i]));

		gwt_plot_curve_set_line_style(GWT_PLOT_CURVE(priv->plot_curve_t[i]), index);
		gwt_plot_curve_set_line_style(GWT_PLOT_CURVE(priv->plot_curve_d[i]), index);
	}
}

void plot_symbol_combo_cb(GtkComboBox *combo, gpointer user_data)
{
	CanPlotDialogPrivate *priv;
	CanPlotDialog* dlg = (CanPlotDialog*)user_data;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	for (i=0; i < NR_PLOTS; i++) {
		int index = gtk_combo_box_get_active(GTK_COMBO_BOX(priv->marker_type[i]));

		gwt_plot_curve_set_symbol(GWT_PLOT_CURVE(priv->plot_curve_t[i]), index);
		gwt_plot_curve_set_symbol(GWT_PLOT_CURVE(priv->plot_curve_d[i]), index);
	}
}


void
create_y_axis_page(CanPlotDialog *dlg, int n)
{
	CanPlotDialogPrivate *priv;
	GtkWidget *label;
	GtkWidget *alignment;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);


	alignment = gtk_alignment_new (0, 0, 0, 0);

	priv->y_axis_page[n] =  gtk_table_new(2, 5, FALSE);

	gtk_container_add (GTK_CONTAINER(alignment), priv->y_axis_page[n]);

	priv->y_axis_page_label[n] = gtk_label_new("Y-Axis Settings");


	/* label */
	label = gtk_label_new("Label:");
	priv->edit_label[n] = gtk_entry_new();

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), label, 0, 1, 0, 1,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), priv->edit_label[n], 1, 2, 0, 1,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	g_signal_connect (G_OBJECT (priv->edit_label[n]), "notify::text", G_CALLBACK (y_axis_label_edit_cb), dlg);


	/* color */
	label = gtk_label_new("Color:");
	priv->color_button[n] = gtk_color_button_new();

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), label, 0, 1, 1, 2,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), priv->color_button[n], 1, 2, 1, 2,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	g_signal_connect (G_OBJECT (priv->color_button[n]), "color-set", G_CALLBACK (color_button_cb), dlg);


/*
	GWT_PLOT_CURVE_STYLE_NONE,
	GWT_PLOT_CURVE_STYLE_FITTED,
	GWT_PLOT_CURVE_STYLE_LINES,
	GWT_PLOT_CURVE_STYLE_STEPS,
	GWT_PLOT_CURVE_STYLE_STICKS,
*/


	/* plot type */
	label = gtk_label_new("Plot Type:");
	priv->plot_type[n] = gtk_combo_box_new_text();
	
	

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), label, 0, 1, 2, 3,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), priv->plot_type[n], 1, 2, 2, 3,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->plot_type[n]), "None" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->plot_type[n]), "Fitted" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->plot_type[n]), "Lines" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->plot_type[n]), "Steps" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->plot_type[n]), "Sticks" );

	g_signal_connect (G_OBJECT (priv->plot_type[n]), "changed", G_CALLBACK (plot_style_combo_cb), dlg);


/*
	GWT_PLOT_CURVE_LINE_STYLE_NORMAL,
	GWT_PLOT_CURVE_LINE_STYLE_DOT,
	GWT_PLOT_CURVE_LINE_STYLE_DOT_DASH,
	GWT_PLOT_CURVE_LINE_STYLE_DASH,
*/

	/* line type */
	label = gtk_label_new("Line Type:");
	priv->line_type[n] = gtk_combo_box_new_text();

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), label, 0, 1, 3, 4,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), priv->line_type[n], 1, 2, 3, 4,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->line_type[n]), "Normal" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->line_type[n]), "Dot" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->line_type[n]), "Dot Dash" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->line_type[n]), "Dash" );

	g_signal_connect (G_OBJECT (priv->line_type[n]), "changed", G_CALLBACK (plot_line_style_combo_cb), dlg);



/*	GWT_PLOT_CURVE_SYMBOL_NONE,
	GWT_PLOT_CURVE_SYMBOL_CROSS,
	GWT_PLOT_CURVE_SYMBOL_XCROSS,
	GWT_PLOT_CURVE_SYMBOL_DOT,
	GWT_PLOT_CURVE_SYMBOL_SQUARE,
	GWT_PLOT_CURVE_SYMBOL_DIAMOND,
*/

	/* Marker type */
	label = gtk_label_new("Marker Type:");
	priv->marker_type[n] = gtk_combo_box_new_text();

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), label, 0, 1, 4, 5,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK,
			0, 0);

	gtk_table_attach(GTK_TABLE (priv->y_axis_page[n]), priv->marker_type[n], 1, 2, 4, 5,
			GTK_SHRINK,
			GTK_SHRINK,
			0, 0);

	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "None" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "Cross" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "X-Cross" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "Dot" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "Square" );
	gtk_combo_box_append_text(GTK_COMBO_BOX(priv->marker_type[n]), "Diamond" );

	g_signal_connect (G_OBJECT (priv->marker_type[n]), "changed", G_CALLBACK (plot_symbol_combo_cb), dlg);

	gtk_notebook_append_page( GTK_NOTEBOOK(priv->notebook), alignment, priv->y_axis_page_label[n]);


}


void set_defaults(CanPlotDialog *dlg) 
{
	CanPlotDialogPrivate *priv;
	int i;

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);


	gtk_entry_set_text(GTK_ENTRY(priv->edit_distance), "Distance [m]");
	gtk_entry_set_text(GTK_ENTRY(priv->edit_time), "Time [s]");

	for (i = 0; i < NR_PLOTS; i++) {
		gtk_entry_set_text(GTK_ENTRY(priv->edit_label[i]), default_y_labels[i]);
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->plot_type[i]), 1);
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->line_type[i]), 0);
		gtk_combo_box_set_active(GTK_COMBO_BOX(priv->marker_type[i]), 2);
		gtk_color_button_set_color(GTK_COLOR_BUTTON(priv->color_button[i]), &default_color[i]);
	}

	color_button_cb(NULL, dlg);
	x_axis_radio_toggle_cb(NULL, dlg);
}


static void
can_plot_dialog_init (CanPlotDialog *dlg)
{
	CanPlotDialogPrivate *priv;
	int i;
	GdkColor c;
	char tmp[128];

	priv = CAN_PLOT_DIALOG_GET_PRIVATE (dlg);

	create_plots(dlg);

	create_plot_frame(dlg);

	priv->vbox = gtk_vbox_new(FALSE, 10);
	priv->notebook = gtk_notebook_new();

	create_x_axis_page(dlg);
	
	for (i = 0; i < NR_PLOTS; i++) {
		create_y_axis_page(dlg, i);
	}

	gtk_box_pack_start(GTK_BOX(priv->vbox), priv->frame, TRUE, TRUE, 10);
	gtk_box_pack_start(GTK_BOX(priv->vbox), priv->notebook, FALSE, FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(priv->vbox), 5);

	gtk_container_add(GTK_CONTAINER(dlg), priv->vbox);
	gtk_widget_show_all(priv->vbox);

	set_defaults(dlg);
}

GtkWidget *
can_plot_dialog_new (void)
{
	return g_object_new (TYPE_CAN_PLOT_DIALOG, NULL);
}

